/****************************************************************************
 * tx.feature.read_scan
 *   A mixin "Read_Scan" that suplies Scanner like feature of TextReaders
 *   Currently, this mixin is the basis of all other Read_ mixins, that is,
 *   all other mixins are replying on hasNextLine/nextLine method.
 *   This is not a good design so should be fixed...
 ****************************************************************************/

module tx.feature.read_scan;
public
{
	import std.regexp;
	import std.conv;
	import std.ctype;
}

template Read_Scan(alias base)
{
public:
	bool hasNextLine()  { return lineBuffers.length ? true : eatToBuffer(); }
	string  nextLine()  { return hasNextLine ? shiftBuffer() : null; }
	string[] allLines() { string[] result; while(hasNextLine) result~=nextLine; return result; }

	bool hasNextInt()   { return hasNext!(int); }
	int     nextInt()   { return getNext!(int); }
	int[]    allInts()  { return allNext!(int); }
	int[]    allIntsInLine() { return allNextInLine!(int); }

	bool hasNextReal()  { return hasNext!(real); }
	real    nextReal()  { return getNext!(real); }
	real[]   allReals() { return allNext!(real); }
	real[]   allRealsInLine() { return allNextInLine!(real); }

	bool hasNextString()  { return hasNext!(string); }
	string  nextString()  { return getNext!(string); }
	string[] allStrings() { return allNext!(string); }
	string[] allStringsInLine() { return allNextInLine!(string); }

private:
	string[] lineBuffers;

	bool eatToBuffer()
	{
		string line = cs.toUTF8( base.readLine().idup );
		if( eof && line.length==0 )
			return false;
		lineBuffers ~= line;
		return true;
	}

	string shiftBuffer()
	{
		auto line = lineBuffers[0];
		lineBuffers = lineBuffers[1..$];
		return line;
	}

private:
	int findNonDelimiter(string s) // todo: should
	{
		int i = 0;
		while( i<s.length && std.ctype.isspace(s[i]) )
			++i;
		return i==s.length ? -1 : i;
	}

private:
	enum FindResult { MATCH, PENDING, UNMATCH };

	template finder(T:int)
	{
		RegExp regex;
		static this() { regex = RegExp(`^(\+|\-)?\d+`); }
		T doCast(string s) { return std.conv.to!(int)(s); }
	}

	template finder(T:string)
	{
		RegExp regex;
		static this() { regex = RegExp(`^[^\s]+`); }
		T doCast(string s) { return s; }
	}

	template finder(T:real)
	{
		RegExp regex;
		static this() { regex = RegExp(`^(\+|\-)?(\d*\.)?\d+(e(\+|\-)?\d+)?`); }
		T doCast(string s) { return std.conv.to!(real)(s); }
	}

	FindResult findInLine(T)( ref string line, out T result, bool discard_found )
	{
		alias finder!(T) fn;

		int i = findNonDelimiter(line);
		if( i == -1 )
			return FindResult.PENDING;

		if( auto match = fn.regex.exec(line[i..$]) )
		{
			result = fn.doCast(match[0]);
			if( discard_found )
				line = line[i+match[0].length..$];
			return FindResult.MATCH;
		}

		return FindResult.UNMATCH;
	}

	bool find(T)( out T result, bool discard_found )
	{
		for(int i=0;; ++i)
		{
			if( i >= lineBuffers.length )
				if( !eatToBuffer() )
					return false;
			switch( findInLine(lineBuffers[i], result, discard_found) )
			{
			case FindResult.MATCH:
				if( discard_found )
					lineBuffers = lineBuffers[i..$];
				return true;
			case FindResult.PENDING:
				break;
			case FindResult.UNMATCH:
				return false;
			}
		}
		return false;
	}

private:
	bool hasNext(T)() { T result; return find(result, false); }
	T    getNext(T)() { T result; find(result, true); return result; }
	T[]  allNext(T)() { T[] result; while(hasNext!(T)) result~=getNext!(T); return result; }
	T[]  allNextInLine(T)() {
		if( lineBuffers.length == 0 )
			if( !eatToBuffer() )
				return null;
		T[] result;
		for(T tmp;;)
			switch( findInLine(lineBuffers[0], tmp, true) )
			{
			case FindResult.MATCH:
				result ~= tmp;
				break;
			case FindResult.PENDING:
			case FindResult.UNMATCH:
				shiftBuffer();
				return result;
			}
		return result;
	}
}
